<?php

namespace Netzmacher\Xblog\DataProcessing;

use Netzmacher\Xblog\Utility\TcaUtility;
use TYPO3\CMS\Core\Database\Connection;
use TYPO3\CMS\Core\Database\ConnectionPool;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Frontend\ContentObject\ContentDataProcessor;
use TYPO3\CMS\Frontend\ContentObject\ContentObjectRenderer;
use TYPO3\CMS\Frontend\ContentObject\DataProcessorInterface;

/*
 * Copyright notice
 *
 * (c) 2020-2021 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * All rights reserved
 *
 * This file is part of the TYPO3 CMS project.
 *
 * It is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, either version 2
 * of the License, or any later version.
 *
 * For the full copyright and license information, please read the
 * LICENSE.txt file that was distributed with this source code.
 *
 * The TYPO3 project - inspiring people to share!
 * 
 * @version 0.20.76
 * @since 0.9.12
 * @internal #t4753 
 */

/**
 * Fetch records from the database, using the default .select syntax from TypoScript.
 *
 * This way, e.g. a FLUIDTEMPLATE cObject can iterate over the array of records.
 *
 * Example TypoScript configuration:
 *
 * 	30 = Netzmacher\Xblog\DataProcessing\DatabaseQueryForeignTableProcessor
 * 	30 {
 * 	  if =
 * 	  if.isTrue.field {
 * 	    cObject = TEXT
 * 	    cObject {
 * 	      data = register:pi1__pi_flexform__settings_flexform_pi_bridge_category
 * 	    }
 * 	  }
 * 
 * 		// local table
 * 	  table = TEXT
 * 	  table.data = register:xBlogPiTable
 * 	
 * 		// field with the relation of the local table. Example: category
 * 	  relationfield = TEXT
 * 	  relationfield.data = register:pi1__pi_flexform__settings_flexform_pi_bridge_category
 * 	
 * 		// uid of the current record of the local table
 * 	  uid = TEXT
 * 	  uid.field = uid
 * 	
 * 		// Significant in case of a relation without an MM table
 * 	  relationvalue = TEXT
 * 	  relationvalue.field {
 * 	    cObject = TEXT
 * 	    cObject {
 * 	      data = register:pi1__pi_flexform__settings_flexform_pi_bridge_category
 * 	    }
 * 	  }
 * 	
 * 	  andWhere = TEXT
 * 	  andWhere = 
 * 	
 * 	  orderBy = TEXT
 * 	  orderBy = 
 * 	
 * 	  as = xBlogFieldCategory
 * 	}
 *
 * where "as" means the variable to be containing the result-set from the DB query.
 */
class DatabaseQueryForeignTableProcessor implements DataProcessorInterface
{

	/**
	 * @var ContentDataProcessor
	 * 
	 * @version 0.9.12
	 * @since 0.9.12
	 */
	protected $contentDataProcessor;

	/**
	 * @var _cObj
	 * 
	 * @version 0.9.12
	 * @since 0.9.12
	 */
	private $_cObj;

	/**
	 * @var _andWhere
	 * 
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private $_andWhere;

	/**
	 * @var _firstResult
	 * 
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private $_firstResult;

	/**
	 * @var _localtable
	 * 
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private $_localtable;

	/**
	 * @var _maxResults
	 * 
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private $_maxResults;

	/**
	 * @var _orderBy
	 * 
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private $_orderBy;

	/**
	 * @var _relationfield
	 * 
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private $_relationfield;

	/**
	 * @var _relationvalue
	 * 
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private $_relationvalue;

	/**
	 * @var _uidlocal
	 * 
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private $_uidlocal;

	/**
	 * Constructor
	 */
	public function __construct()
	{
		$this->contentDataProcessor = GeneralUtility::makeInstance( ContentDataProcessor::class );
	}

	/**
	 * Fetches records from the database as an array
	 *
	 * @param ContentObjectRenderer $cObj The data of the content element or page
	 * @param array $contentObjectConfiguration The configuration of Content Object
	 * @param array $processorConfiguration The configuration of this processor
	 * @param array $processedData Key/value store of processed data (e.g. to be passed to a Fluid View)
	 *
	 * @return array the processed data as key/value store
	 * @version 0.9.12
	 * @since 0.9.12
	 */
	public function process( ContentObjectRenderer $cObj, array $contentObjectConfiguration, array $processorConfiguration, array $processedData )
	{
		// Isn't used
		unset( $contentObjectConfiguration );

		if( isset( $processorConfiguration[ 'if.' ] ) && !$cObj->checkIf( $processorConfiguration[ 'if.' ] ) )
		{
			return $processedData;
		}

		$this->_init( $cObj, $processorConfiguration );

		$this->_localtable = $this->_cObjGetSingle( 'table' );
		$this->_relationfield = $this->_cObjGetSingle( 'relationfield' );
		$this->_relationvalue = $this->_cObjGetSingle( 'relationvalue' );
		$this->_uidlocal = $this->_cObjGetSingle( 'uid' );
		$this->_andWhere = $this->_cObjGetSingle( 'andWhere' );
		$this->_orderBy = $this->_cObjGetSingle( 'orderBy' );
		$this->_firstResult = $this->_cObjGetSingle( 'firstResult' );
		$this->_maxResults = $this->_cObjGetSingle( 'maxResults' );
//		var_dump( __METHOD__, __LINE__, $this->_pConf);
//		var_dump( __METHOD__, __LINE__, $this->_localtable, $this->_relationfield, $this->_relationvalue, $this->_uidlocal, $this->_andWhere, $this->_orderBy );
//		var_dump( __METHOD__, __LINE__, $GLOBALS['TSFE']->register );
//		var_dump( __METHOD__, __LINE__, $this->_orderBy );
//		var_dump( __METHOD__, __LINE__, $this->_andWhere );



		if( !$this->_requirements() )
		{
			return $processedData;
		}

		//var_dump( __METHOD__, __LINE__,  );
		// The variable to be used within the result
		$targetVariableName = $cObj->stdWrapValue( 'as', $processorConfiguration, 'records' );

		$foreigntable = TcaUtility::getColumnConfigForeigntable( $this->_localtable, $this->_relationfield );
		//var_dump( __METHOD__, __LINE__, $foreigntable );
		// Execute a SQL statement to fetch the records
		$rows = $this->_rowsByRelation();
		//var_dump( __METHOD__, __LINE__, $rows );
		$processedRecordVariables = [];
		foreach( ( array ) $rows as $key => $record )
		{
			/** @var ContentObjectRenderer $recordContentObjectRenderer */
			$recordContentObjectRenderer = GeneralUtility::makeInstance( ContentObjectRenderer::class );
			$recordContentObjectRenderer->start( $record, $foreigntable );
			$processedRecordVariables[ $key ] = [ 'data' => $record ];
			$processedRecordVariables[ $key ] = $this->contentDataProcessor->process( $recordContentObjectRenderer, $processorConfiguration, $processedRecordVariables[ $key ] );
		}

		$processedData[ $targetVariableName ] = $processedRecordVariables;

//		var_dump( __METHOD__, __LINE__, $processedData );
//		die();
		return $processedData;
	}

	/**
	 * _cObjGetSingle
	 *
	 * @param string	$element: name of the element of the $processorConfiguration
	 *
	 * @return string 
	 * @version 0.9.12
	 * @since 0.9.12
	 */
	private function _cObjGetSingle( $element )
	{
		$conf = $this->_pConf;

		return $this->_cObj->cObjGetSingle( $conf[ $element ], $conf[ $element . '.' ] );
	}

	/**
	 * _init
	 *
	 * @param ContentObjectRenderer $cObj The data of the content element or page
	 * @param array $processorConfiguration The configuration of this processor
	 *
	 * @return string 
	 * @version 0.9.12
	 * @since 0.9.12
	 */
	private function _init( $cObj, $processorConfiguration )
	{
		$this->_cObj = $cObj;
		$this->_pConf = $processorConfiguration;
	}

	/**
	 * _isMM
	 * 
	 * @return	boolean
	 * @version 0.30.19
	 * @since		0.9.12
	 */
	private function _isMM()
	{
		return TcaUtility::isMM( $this->_localtable, $this->_relationfield );
	}

	/**
	 * _isTypeInline
	 * 
	 * @return	boolean
	 * @version 0.30.19
	 * @since		0.9.12
	 */
	private function _isTypeInline()
	{
		return TcaUtility::isTypeInline( $this->_localtable, $this->_relationfield );
	}

	/**
	 * _queryPartAndWhere
	 * 
	 * @return array	$processedData
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private function _queryPartAndWhere( $queryBuilder )
	{
		if( !$this->_andWhere )
		{
			return $queryBuilder;
		}

		$queryBuilder->andWhere( $this->_andWhere );
		return $queryBuilder;
	}

	/**
	 * _queryPartFirstResult
	 * 
	 * @return array	$processedData
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private function _queryPartFirstResult( $queryBuilder )
	{
		if( !$this->_firstResult )
		{
			return $queryBuilder;
		}

		$queryBuilder->setFirstResult( ( int ) $this->_firstResult );

		return $queryBuilder;
	}

	/**
	 * _queryPartMaxResults
	 * 
	 * @return array	$processedData
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private function _queryPartMaxResults( $queryBuilder )
	{
		if( !$this->_maxResults )
		{
			return $queryBuilder;
		}

		$queryBuilder->setMaxResults( ( int ) $this->_maxResults );

		return $queryBuilder;
	}

	/**
	 * _queryPartOrderBy
	 * 
	 * @return array	$processedData
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private function _queryPartOrderBy( $queryBuilder )
	{
		if( !$this->_orderBy )
		{
			return $queryBuilder;
		}

		$statements = explode( '||', $this->_orderBy );
		foreach( $statements as $statement )
		{
			list($field, $order) = explode( '|', $statement );
			if( empty( $order ) )
			{
				$order = 'ASC';
			}
			$order = strtoupper( $order );
			$queryBuilder->addOrderBy( $field, $order );
		}

		return $queryBuilder;
	}

	/**
	 * _queryParts
	 * 
	 * @return array	$processedData
	 * @version 0.9.15
	 * @since 0.9.15
	 */
	private function _queryParts( $queryBuilder )
	{
		$queryBuilder = $this->_queryPartAndWhere( $queryBuilder );
		$queryBuilder = $this->_queryPartOrderBy( $queryBuilder );
		$queryBuilder = $this->_queryPartFirstResult( $queryBuilder );
		$queryBuilder = $this->_queryPartMaxResults( $queryBuilder );
		return $queryBuilder;
	}

	/**
	 * _requirements
	 *
	 * @return boolean
	 * @version 0.9.12
	 * @since 0.9.12
	 */
	private function _requirements()
	{
		switch( true )
		{
			case( empty( $this->_localtable ) ):
			case( empty( $this->_relationfield ) ):
			case( empty( $this->_uidlocal ) ):
				return false;
			default:
				return true;
		}
	}

	/**
	 * _rowsByRelation
	 * 
	 * @return array	$rows
	 * @version 0.9.14
	 * @since 0.9.12
	 * @internal #t4753 
	 */
	private function _rowsByRelation()
	{
		switch( true )
		{
			case( $this->_isTypeInline()):
				$rows = $this->_rowsByRelationTypeInline();
				break;
			case( $this->_isMM()):
				$rows = $this->_rowsByRelationMM();
				break;
			default:
				$rows = $this->_rowsByRelationDefault();
				break;
		}

		return $rows;
	}

	/**
	 * _rowsByRelationDefault
	 * 
	 * _rowsByRelation
	 * 
	 * @return array	$processedData
	 * @version 0.9.15
	 * @since 0.9.12
	 */
	private function _rowsByRelationDefault()
	{
		$foreigntable = TcaUtility::getColumnConfigForeigntable( $this->_localtable, $this->_relationfield );

		$queryBuilder = GeneralUtility::makeInstance( ConnectionPool::class )->getQueryBuilderForTable( $foreigntable );
		$queryBuilder
						->select( '*' )
						->from( $foreigntable )
						->where(
										$queryBuilder->expr()->inSet( 'uid', $queryBuilder->createNamedParameter( $this->_relationvalue ) )
						)
		;
		$queryBuilder = $this->_queryParts( $queryBuilder );

		$rows = $queryBuilder->execute()->fetchAll();

		return $rows;
	}

	/**
	 * _rowsByRelationMM
	 * 
	 * @return array	$processedData
	 * @version 0.20.68
	 * @since 0.9.12
	 * @internal #t4753 
	 */
	private function _rowsByRelationMM()
	{
		$MMtable = TcaUtility::getColumnConfigMM( $this->_localtable, $this->_relationfield );

		switch( true )
		{
			case($MMtable == 'sys_category_record_mm'):
				$rows = $this->_rowsByRelationMMsyscategory( $MMtable );
				break;
			case($MMtable != 'sys_category_record_mm'):
			default:
				$rows = $this->_rowsByRelationMMdefault();
				break;
		}

		return $rows;
	}

	/**
	 * _rowsByRelationMM
	 * 
	 * @param string $MMtable
	 * 
	 * @return array	$processedData
	 * @version 0.20.68
	 * @since 0.20.68
	 */
	private function _rowsByRelationMMsyscategory( $MMtable )
	{
		$foreigntable = TcaUtility::getColumnConfigForeigntable( $this->_localtable, $this->_relationfield );

		$MMmatchfieldsForeign = $this->_localtable;
		$MMmatchfieldsLocal = 'sys_category';

		$queryBuilder = GeneralUtility::makeInstance( ConnectionPool::class )->getQueryBuilderForTable( $MMtable );
		$queryBuilder
						->select( '*' )
						->from( $MMtable )
						->where(
										$queryBuilder->expr()->eq( 'uid_foreign', $queryBuilder->createNamedParameter( $this->_uidlocal ) )
						)
		;

		if( $MMmatchfieldsLocal )
		{
			$queryBuilder
							->andWhere(
											$queryBuilder->expr()->eq( 'fieldname', $queryBuilder->createNamedParameter( $MMmatchfieldsLocal ) )
											, $queryBuilder->expr()->eq( 'tablenames', $queryBuilder->createNamedParameter( $MMmatchfieldsForeign ) )
							)
			;
		}

		$queryBuilder = $this->_queryParts( $queryBuilder );

		$statement = $queryBuilder->execute();

//		var_dump( __METHOD__, __LINE__, $queryBuilder->getSQL(), $queryBuilder->getParameters() );
//		die();


		while( $row = $statement->fetch() )
		{
			$aUidForeign[] = $row[ 'uid_local' ];
		}
		//var_dump( __METHOD__, __LINE__, $aUidForeign, $row );

		$queryBuilder = GeneralUtility::makeInstance( ConnectionPool::class )->getQueryBuilderForTable( $foreigntable );
		$queryBuilder
						->select( '*' )
						->from( $foreigntable )
						->where(
										$queryBuilder->expr()->in( 'uid', $queryBuilder->createNamedParameter(
																		( array ) $aUidForeign, Connection::PARAM_INT_ARRAY
														)
										)
						)
		;

		$rows = $queryBuilder->execute()->fetchAll();

//		var_dump( __METHOD__, __LINE__, $queryBuilder->getSQL(), $queryBuilder->getParameters(), $rows );
//		die();
//		var_dump( __METHOD__, __LINE__, $foreigntable, $aUidForeign, $rows );
		return $rows;
	}

	/**
	 * _rowsByRelationMM
	 * 
	 * @return array	$processedData
	 * @version 0.20.76
	 * @since 0.9.12
	 * @internal #t4753 
	 */
	private function _rowsByRelationMMdefault()
	{
		$foreigntable = TcaUtility::getColumnConfigForeigntable( $this->_localtable, $this->_relationfield );
		$MMtable = TcaUtility::getColumnConfigMM( $this->_localtable, $this->_relationfield );
		// Isn't used
		//$MMmatchfields = TcaUtility::getColumnConfigMMmatchfields(  );
		$MMmatchfieldsForeign = TcaUtility::getColumnConfigMMmatchfieldsForeign( $this->_localtable, $this->_relationfield );
		$MMmatchfieldsLocal = TcaUtility::getColumnConfigMMmatchfieldsLocal( $this->_localtable, $this->_relationfield );

		$this->_rowsByRelationMMdefaultOpposite();

		$queryBuilder = GeneralUtility::makeInstance( ConnectionPool::class )->getQueryBuilderForTable( $MMtable );
		$queryBuilder
						->select( '*' )
						->from( $MMtable )
						->where(
										$queryBuilder->expr()->eq( 'uid_local', $queryBuilder->createNamedParameter( $this->_uidlocal ) )
						)
		;

		if( $MMmatchfieldsLocal )
		{
			$queryBuilder
							->andWhere(
											$queryBuilder->expr()->eq( 'table_local', $queryBuilder->createNamedParameter( $MMmatchfieldsLocal ) ), $queryBuilder->expr()->eq( 'table_foreign', $queryBuilder->createNamedParameter( $MMmatchfieldsForeign ) )
							)
			;
		}

		$queryBuilder = $this->_queryParts( $queryBuilder );

		$statement = $queryBuilder->execute();

//		var_dump( __METHOD__, __LINE__, $queryBuilder->getSQL(), $queryBuilder->getParameters() );
//		die();


		while( $row = $statement->fetch() )
		{
			$aUidForeign[] = $row[ 'uid_foreign' ];
		}
//		var_dump( __METHOD__, __LINE__, $aUidForeign, $row );

		$queryBuilder = GeneralUtility::makeInstance( ConnectionPool::class )->getQueryBuilderForTable( $foreigntable );
		$queryBuilder
						->select( '*' )
						->from( $foreigntable )
						->where(
										$queryBuilder->expr()->in( 'uid', $queryBuilder->createNamedParameter(
																		( array ) $aUidForeign, Connection::PARAM_INT_ARRAY
														)
										)
						)
		;

		$rows = $queryBuilder->execute()->fetchAll();

//		var_dump( __METHOD__, __LINE__, $queryBuilder->getSQL(), $queryBuilder->getParameters(), $rows );
//		die();
//		var_dump( __METHOD__, __LINE__, $foreigntable, $aUidForeign, $rows );
		return $rows;
	}

	/**
	 * _rowsByRelationMMdefaultOpposite
	 * 
	 * @return void
	 * @version 0.20.76
	 * @since 0.20.76
	 */
	private function _rowsByRelationMMdefaultOpposite()
	{
		$MMoppositeField = TcaUtility::getColumnConfigMMoppositeField( $this->_localtable, $this->_relationfield );
		if( !$MMoppositeField )
		{
			return;
		}

		$prompt = 'Sorry, relations with an opposite field aren\'t supported :( This is a feature request ;) '
						. __METHOD__ . '#' . __LINE__;
		die( $prompt );
	}

	/**
	 * _rowsByRelationTypeInline
	 * 
	 * @return array	$processedData
	 * @version 0.9.15
	 * @since 0.9.14
	 */
	private function _rowsByRelationTypeInline()
	{
		$foreignfield = TcaUtility::getColumnConfigForeignfield( $this->_localtable, $this->_relationfield );
		$foreigntable = TcaUtility::getColumnConfigForeigntable( $this->_localtable, $this->_relationfield );

		$queryBuilder = GeneralUtility::makeInstance( ConnectionPool::class )->getQueryBuilderForTable( $foreigntable );
		$queryBuilder
						->select( '*' )
						->from( $foreigntable )
						->where(
										$queryBuilder->expr()->inSet( $foreignfield, $queryBuilder->createNamedParameter( $this->_uidlocal ) )
						)
		;
		$queryBuilder = $this->_queryParts( $queryBuilder );

//		debug($queryBuilder->getSQL());
//		debug($queryBuilder->getParameters());


		$rows = $queryBuilder->execute()->fetchAll();

		return $rows;
	}

}
